/* exhaust.c: crippled pmars-like redcode simulator
 * $Id: exhaust.c,v 1.8 2000/08/15 00:25:16 rowan Exp $
 */

/* This file is part of `exhaust', a memory array redcode simulator.
 * Copyright (C) 2000 M Joonas Pihlaja
 *    
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * You may contact the author by:
 *   email: jpihlaja@cc.helsinki.fi
 *   mail:  M Joonas Pihlaja
 *          Artturi Kannistontie 4A11
 *          00320, Helsinki
 *          Finland
 */

#define VERSION 1
#define REVISION 7

/* Note pmars incompatibility: Minimum warrior separation (minsep) is
   handled wrong by pmars (prior to 0.8.6) when minsep < MAXLENGTH.
   This affects running warriors in small cores since the other
   warriors might be loaded outside of core (it might even crash).
*/
#include <stdio.h>
#ifdef SYSV
#include <strings.h>
#else
#include <string.h>
#endif
#include <stdlib.h>
#include <time.h>
#include <ctype.h>

#include "exhaust.h"		/* types */
#include "asm.h"		/* assembler proto */
#include "sim.h"		/* simulator proto */


/* Globals to communicate options from readargs() to main() */
int OPT_rounds = 1;		/* number of rounds to fight */
int OPT_cycles = 80000;		/* cycles until tie */
unsigned int OPT_coresize = 8000; /* default core size */
int OPT_minsep = 0;		/* minimum distance between warriors */
int OPT_processes = 8000;	/* default max. procs./warrior */
char *OPT_w1_name;		/* filename of warrior 1 */
char *OPT_w2_name;		/* filename of warrior 2 */
int OPT_F = 0;			/* initial position of warrior 2 */
int OPT_k = 0;			/* nothing (`koth' format flag) */
int OPT_b = 0;			/* nothing (we are always `brief') */

char *prog_name;
char errmsg[256];

#define NPOS ( OPT_coresize+1 - 2*OPT_minsep )

#define min(x,y) ((x)<(y) ? (x) : (y))

/* xbasename(): custom basename(3); returns the filename from a path
 */
static
char *
xbasename(s)
     char *s;
{
  char *b;
  b = strrchr(s,'/');
  return b ? 1+b : s;
}

void
panic( char *msg )
{
  fprintf(stderr, "%s: ", prog_name );
  fprintf(stderr, "%s", msg );
  exit(1);
}


/* pmars/rng.c random number generator
 */
s32_t
rng(seed)
  s32_t   seed;
{
  register s32_t temp = seed;
  temp = 16807 * (temp % 127773) - 2836 * (temp / 127773);
  if (temp < 0)
    temp += 2147483647;
  return temp;
}


/* stub to test that the assembled warrior is OK */
void
init_warrior( w )
     warrior_t *w;
{
  if ( w->len == 0 ) {
    sprintf(errmsg, "no lines in warrior %d '%s'\n", w->no, w->name);
    panic(errmsg);
  }
}

void
usage(prog)
     char *prog;
{
  printf("%s v%d.%d\n", prog_name, VERSION, REVISION );
  printf("usage: %s [opts] war1 war2\n", prog_name );
  printf("opts: -r <rounds>, -c <cycles>, -F <pos>, -s <coresize>,\n"
	  "      -p <maxprocs>, -d <minsep>, -bk\n");
  exit(1);
}

/*
 * parse options
 */
void
readargs( argc, argv )
     int argc;
     char **argv;
{
  int n;
  char c;
  int cix;

  OPT_w1_name = OPT_w2_name = NULL;

  n = 1;
  while ( n < argc ) {
    cix = 0;
    c = argv[n][cix++];
    if ( c == '-' && argv[n][1] ) {
      do {
	c = argv[n][cix++];
	if (c)
	  switch (c) {

	  case 'k': OPT_k = 1; break;
	  case 'b': OPT_b = 1; break;

	  case 'F':
	    if ( n == argc-1 || !isdigit(argv[n+1][0]) )
	      panic( "bad argument for option -F\n");
	    c = 0;
	    OPT_F = atoi( argv[++n] );
	    break;

	  case 's':
	    if ( n == argc-1 || !isdigit(argv[n+1][0]) )
	      panic( "bad argument for option -s\n");
	    c = 0;
	    OPT_coresize = atoi( argv[++n] );
	    if ( OPT_coresize <= 0 )
	      panic( "core size must be > 0\n");
	    break;

	  case 'd':
	    if ( n == argc-1 || !isdigit(argv[n+1][0]) )
	      panic( "bad argument for option -d\n");
	    c = 0;
	    OPT_minsep = atoi( argv[++n] );
	    if ( (int)OPT_minsep <= 0 )
	      panic( "minimum warrior separation must be > 0\n" );
	    break;

	  case 'p':
	    if ( n == argc-1 || !isdigit(argv[n+1][0]) )
	      panic( "bad argument for option -p\n");
	    c = 0;
	    OPT_processes = atoi( argv[++n] );
	    if ( OPT_processes <= 0 )
	      panic( "max processes must be > 0\n" );
	    break;

	  case 'r':
	    if ( n == argc-1 || !isdigit(argv[n+1][0]) )
	      panic( "bad argument for option -r\n");
	    c = 0;
	    OPT_rounds = atoi( argv[++n] );
	    if ( OPT_rounds < 0 )
	      panic( "can't do a negative number of rounds!\n" );
	    break;
	  case 'c':
	    if ( n == argc-1 || !isdigit(argv[n+1][0]) )
	      panic( "bad argument for option -c\n");
	    c = 0;
	    OPT_cycles = atoi( argv[++n] );
	    if ( OPT_cycles <= 0 )
	      panic( "cycles must be > 0\n" );
	    break;
	  default:
	    sprintf(errmsg,"unknown option '%c'\n", c);
	    panic(errmsg);
	  }
      } while (c);

    } else /* it's a file name */ {
      if ( OPT_w1_name )
	OPT_w2_name = argv[n];
      else
	OPT_w1_name = argv[n];
    }
    n++;
  }

  
  /* A few sanity checks */
  if ( OPT_minsep == 0 ) OPT_minsep = min( OPT_coresize/2, MAXLENGTH );
  if ( (int)NPOS <= 0 ) {
    sprintf(errmsg,  "no possible starting positions! (minimum distance"
	   " (=%d) too large?)\n", OPT_minsep);
    panic( errmsg );
  }

  if ( OPT_F )
    if ( OPT_F > OPT_minsep + NPOS || OPT_F < OPT_minsep ) {
      sprintf( errmsg,
	      "F number must be inside [%d, %d)\n",
	      OPT_minsep, OPT_minsep+NPOS );
      panic( errmsg );
    }

  if (! OPT_w1_name )
    usage(argv[0]);
}


/*-------------------------------------------------------------------------
 * Main
 */

int
main( int argc, char **argv )
{
  warrior_t w1, w2;
  int res_tab[3] = {0, 0, 0};	/* result table: {w1 wins, w2 wins, ties} */
  field_t start1, start2;	/* load positions of w1 and w2 */
  int res, n;			/* temps */
  int pos, nwar;		/* temps */
  s32_t seed;
  insn_t *core;			/* ptr. to core memory */

  prog_name = xbasename(argv[0]);
  readargs(argc, argv);
  seed = OPT_F ? OPT_F-OPT_minsep : rng(time(0)*0x1d872b41);

  /*
   * Assemble the warriors
   */
  memset(&w1,0,sizeof(w1));
  nwar = 1;
  asm_fname( OPT_w1_name, &w1, OPT_coresize );
  w1.name = OPT_w1_name; w1.no = 1;
  init_warrior( &w1 );

  w2.len = 0;
  if ( OPT_w2_name ) {
    ++nwar;
    memset(&w2,0,sizeof(w2));
    asm_fname( OPT_w2_name, &w2, OPT_coresize );
    w2.name = OPT_w2_name; w2.no = 2;
    init_warrior( &w2 );
  }

  /* More sanity checks */
  if ( nwar > 1 && 
       ( OPT_minsep < w1.len
	 || OPT_minsep < w2.len ))
    panic( "minimum separation must be >= warrior length(s)\n" );

  if ( OPT_coresize < w1.len + w2.len )
    panic( "core too small for warrior(s)!\n" );


  /*
   * Allocate core memory
   */
  core = sim_alloc_bufs( nwar, OPT_coresize, OPT_processes, OPT_cycles );
  if ( core == NULL )
    panic("can't allocate memory.\n");

  /*
   * Fight OPT_rounds rounds.
   */
  start1 = w1.start;
  for ( n = 0; n < OPT_rounds; n++ ) {
    memset(core+w1.len, 0, sizeof(insn_t)*(OPT_coresize-w1.len));
    pos = OPT_minsep + seed % NPOS;
    seed = rng(seed);

    /* Load warrior(s) into core. */
    memcpy( core, w1.code, sizeof(insn_t)*w1.len );
    if ( nwar == 2 )
      memcpy( core+pos, w2.code, sizeof(insn_t)*w2.len );

    /* Compute starting positions of warriors. */
    /* Swap start1 and start2 on an odd round */
    /*   to make the second warrior start first. */
    if ( nwar == 2 && (n & 1) == 1 ) {
      start1 = (w2.start+pos) % OPT_coresize;
      start2 = w1.start;
    } else {
      start1 = w1.start;
      start2 = (w2.start+pos) % OPT_coresize;
    }

    res = sim( nwar, start1, start2, OPT_cycles, NULL );
    if ( res == -1 ) panic( "simulator panic!\n" );
    /*
     * Accumulate result
     */
    if ( nwar == 1 || (n & 1) == 0 ) /* single warrior, or even round */
      ++res_tab[res];
    else /* odd round */
      if (res == 2)
	++res_tab[ 2 ];
      else
	++res_tab[ res ^ 1 ];
  }
  sim_free_bufs();


  /* 
   * Output results `koth' style:
   *
   * For two warriors:
   *   W1_wins W1_ties
   *   W2_wins W2_ties
   *
   * For one warrior:
   *   W1_wins W1_losses
   */
  if ( nwar == 1 ) 
    printf("%d %d\n", res_tab[0], res_tab[1] );
  else {
    printf("%d %d\n", res_tab[0], res_tab[2] );
    printf("%d %d\n", res_tab[1], res_tab[2] );
  }


  return 0;
}
