/* This file is part of 'yace', a corewar-warrior evolver.
 *
 * Copyright (C) 2000 Jason Boer
 * modified 2000 Martin Ankerl <martinankerl@web.de>
 *    
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
 * 02111-1307, USA.
 */

#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <stdlib.h>
#include <time.h>
#include "insn.h"
#include "exhaust.h"
#include "sim.h"
#include "asm.h"

/* #define DEBUG_YACE */

typedef struct warrior_score_st {
    int score;		
    unsigned int fights;	/* token code */
} warrior_score_t;

/********************/
/*these are set from the config file*/

static long population_size=0;
static long number_of_battles=0;

static long OPT_coresize = 8000; /* default core size */
static long OPT_cycle = 80000;          /* cycles until tie */
static long OPT_minsep=100;             /* minimum distance between warriors */
static long OPT_processes = 8000;       /* default max. procs./warrior */

static double mutation_rate= 0.04;
static double insertion_rate=0.05;
static double removal_rate=0.05;

static long max_number_size=0;
static long max_instructions=0;
static long startlength=5;

static long testset_size=3;
static long hill_size=3;
static long current_cycle=0;
static long cycles_before_save=50;
static long cycles_before_status=50;

static char path_symbol[32]="";

/********************/

insn_t *core;           /* address of core */

/* prototypes *******/


char * decode_line(const insn_t * codeline);
    
void setup(warrior_t ** warriors, warrior_score_t ** warriors_score, int ** warriors_selection);
int buffer2double(const char* buffer, const char* searchstring, double *number);
int buffer2long(const char* buffer, const char* searchstring, long *number);
int buffer2string(const char* buffer, const char* searchstring, char* target);

void mars(const warrior_t *war1, const warrior_t *war2, int *w1_score, int *w2_score, int rounds, unsigned long cycles, unsigned long distance);
void create_seed_population(warrior_t * warriors, int population_size, int linecount);

/* fights every warrior against a random taken */
void selection_random(int * warriors_selection, int anz);

void fitness_standard(const int * warriors_selection, const warrior_t * warriors, warrior_score_t * warriors_score, int anz);
void fitness_testset(const int * warriors_selection, const warrior_t * warriors, warrior_score_t * warriors_score, int anz);

void reproduce_standard(const int * warriors_selection, warrior_t * warriors, warrior_score_t * warriors_score, int anz);
void status(const int * warriors_selection, const warrior_t * warriors, warrior_score_t * warriors_score, int anz, int current_cycle);

void save_warriors(const warrior_t * warriors, int anz, int current_cycle);

void straight_replace(const warrior_t * sw, warrior_t * tw);
void create_successor(const warrior_t * sw, warrior_t * tw);

int test_viable(const warrior_t * test_warrior);

int propability(double prop);

int random_corenumber(int range, int numbertype);
int random_number(int range);

void new_line_of_code(insn_t * codeline);
void evolve_a_line(insn_t * codeline);



int main();


/********************************************/
/* end prototypes                           */


/* numbertype 0: small number
   numbertype 1: big number
   numbertype 2: small or big number (random) */
int random_corenumber(int range, int numbertype) {
    int number = range-(random_number(range*2));

    /* random number mode */
    if (numbertype == 2) numbertype = random_number(2); 
    switch(numbertype) {
    case 0: break;
    case 1: number += OPT_coresize/(random_number(3) + 2); break;
    }
    return (MODS(number, OPT_coresize));
}

int random_number(int range) {
    return( ( ((double)(rand())) / RAND_MAX ) *range );
}

/********************************************/

int propability(double prop) {
    double nr;
    nr = (double)(rand()) / RAND_MAX;
    return (nr < prop);
}

void copy(const warrior_t * source, warrior_t *target) {
    *target = *source;
}
  

/********************************************/

int buffer2double(const char* buffer, const char* searchstring, double *number) {
    if(strstr(buffer, searchstring)!=0)     {
        char * parm_ptr=strstr(buffer,"=");
        if (parm_ptr!=0)                {
            *number=atof(++parm_ptr);
            return 1;
        }
    }
    return 0;
}

int buffer2long(const char* buffer, const char* searchstring, long *number) {
    if(strstr(buffer, searchstring)!=0)     {
        char * parm_ptr=strstr(buffer,"=");
        if (parm_ptr!=0)                {
            *number=atof(++parm_ptr);
            return 1;
        }
    }
    return 0;
}

int buffer2string(const char* buffer, const char* searchstring, char* target) {
    if(strstr(buffer, searchstring)!=0)     {
        char * parm_ptr=strstr(buffer,"=");
        if (parm_ptr!=0){
            /* WARNING! only 2 letters are copied!! */
            strncpy(target,++parm_ptr,2);
            return 1;
        }
    }
    return 0;
}

/********************************************/

void setup(warrior_t ** warriors, warrior_score_t ** warriors_score, int ** warriors_selection) {
    FILE * config_file=0;
    char buffer[256];
        
    /*seed the random numbers*/
#ifdef DEBUG_YACE
    srand(967994294);
#else
    time_t t;       
    int nr;    
    nr = (unsigned int) time(& t);
    printf("srand [%d]\n", nr);
    srand(nr);
#endif

    /*set defaults*/
    population_size=26;
    max_instructions=10;
    startlength=5;
    max_number_size=10;
    mutation_rate=25;
    insertion_rate=5;
    removal_rate=5;
    number_of_battles=5;

    OPT_coresize=8000;
    OPT_cycle=80000;
    OPT_minsep=100;
    OPT_processes=8000;

    cycles_before_status=20;
    cycles_before_save=20;

    strcpy(path_symbol,"\\");

    /*read config file if there*/
    if (0==(config_file=fopen("yace.cfg","r")))
    {
        printf("yace.cfg not found using defaults.\n");
    }
    else
    {
        while (fgets(buffer,255,config_file)!=0)
        {                
            /*ignore ; line*/
            if(strstr(buffer,";")!=0) continue;
            if (buffer2long(buffer, "population_size", &population_size)) continue;
            if (buffer2long(buffer, "max_instructions", &max_instructions)) continue;
            if (buffer2long(buffer, "max_number_size", &max_number_size)) continue;
	    if (buffer2long(buffer, "startlength", &startlength)) continue;

            if (buffer2double(buffer, "mutation_rate", &mutation_rate)) continue;
            if (buffer2double(buffer, "insertion_rate", &insertion_rate)) continue;
            if (buffer2double(buffer, "removal_rate", &removal_rate)) continue;

            if (buffer2long(buffer, "number_of_battles", &number_of_battles)) continue;
            
            if (buffer2long(buffer, "OPT_coresize", &OPT_coresize)) continue;
            if (buffer2long(buffer, "OPT_cycle", &OPT_cycle)) continue;
            if (buffer2long(buffer, "OPT_minsep", &OPT_minsep)) continue;
            if (buffer2long(buffer, "OPT_processes", &OPT_processes)) continue;
            
            if (buffer2long(buffer, "testset_size", &testset_size)) continue;
            if (buffer2long(buffer, "hill_size", &hill_size)) continue;

            if (buffer2string(buffer, "path_symbol", path_symbol)) continue;
            if (buffer2long(buffer, "cycles_before_status", &cycles_before_status)) continue;
            if (buffer2long(buffer, "cycles_before_save", &cycles_before_save)) continue;
        }/*while*/
        fclose(config_file);
    }
    
    /* popsize should be dividable by 2! */
    if (population_size % 2) population_size++;

    /* allocate whole memory */
    *warriors = malloc(sizeof(warrior_t)*population_size);
    *warriors_score = malloc(sizeof(warrior_score_t)*population_size);
    *warriors_selection = malloc(sizeof(int)*population_size);
    
    /* request address of core memory */
    core = sim_alloc_bufs( 2, OPT_coresize, OPT_processes, OPT_cycle);
}



void new_line_of_code(insn_t * codeline) {
    int op, m, ma, mb;

    codeline->a = random_corenumber(max_number_size, 2);
    codeline->b = random_corenumber(max_number_size, 2);
    
    op = random_number(TOK_OPCODE_STOP - TOK_OPCODE_START - 1);
    m =  random_number(TOK_MODIFIER_STOP - TOK_MODIFIER_START - 1);
    ma = random_number(TOK_MODE_STOP - TOK_MODE_START - 1);
    mb = random_number(TOK_MODE_STOP - TOK_MODE_START - 1);

    codeline->in = OP(op, m, ma, mb);
}

/********************************************/

int test_viable(const warrior_t * test_warrior) {
    int s1=0;
    int s2=0;
    static warrior_t plant_warrior;
    static int plant_setup=0;

    /* set up plant */
    if (plant_setup == 0) {
	plant_warrior.code[0].in = OP(SPL, mF, DIRECT, DIRECT);
	plant_warrior.code[0].a = 0;
	plant_warrior.code[0].b = 0;
	plant_warrior.len = 1;
	plant_warrior.start=0;

	plant_setup=1;
    }
    
    
    /*test with plant*/
    mars(&plant_warrior, test_warrior, &s1, &s2, number_of_battles, OPT_cycle, OPT_minsep);

    if (s2<1) return 0;
    return 1;

} /*test_viable()*/

/********************************************/

void create_seed_population(warrior_t * warriors, int population_size, int linecount) {
    FILE * warrior_file;
    int i, j;
    char file_name[256];

    /* test linecout */
    if (linecount > max_instructions) linecount=max_instructions;
    else if (linecount < 1) linecount=1;

    
    for (i=0; i < population_size; i++) {
	/* test if saved warrior exists */
	sprintf(file_name, "..%ssave%spop%05d.red", path_symbol, path_symbol, i);
	if (!(warrior_file = fopen(file_name, "r"))) {
	    /* create random warrior */
	    do {
		for (j=0; j<linecount; j++) {
		    new_line_of_code(&warriors[i].code[j]);
		}
		warriors[i].len = linecount;
		warriors[i].start = 0;
	    } while (test_viable(&warriors[i]));
	    printf("Warrior %5d created\n", i);
	}
	else {
	    /* load warrior from disk */
	    asm_fname(file_name, &warriors[i], OPT_coresize);
	    printf("Warrior %5d loaded from disk\n", i);
	}
    }
}/*create_seed_population()*/

/********************************************/

void evolve_a_line(insn_t * codeline) {
    int o, m, ma, mb;
    
    /* opcode */
    if (propability(mutation_rate)) o = random_number(TOK_OPCODE_STOP - TOK_OPCODE_START - 1);
    else o = (codeline->in >> opPOS) & opMASK;

    /* modifiers */
    if (propability(mutation_rate))  m = random_number(TOK_MODIFIER_STOP - TOK_MODIFIER_START - 1);
    else m = (codeline->in >> moPOS) & moMASK;

    /* address mode a */
    if (propability(mutation_rate)) ma = random_number(TOK_MODE_STOP - TOK_MODE_START - 1);
    else ma = (codeline->in >> maPOS) & mMASK;

    /* address mode b */
    if (propability(mutation_rate)) mb = random_number(TOK_MODE_STOP - TOK_MODE_START - 1);
    else mb = (codeline->in >> mbPOS) & mMASK;
    
    /* number a */
    if (propability(mutation_rate)) codeline->a = random_corenumber(max_number_size, 2);

    /* number b */
    if (propability(mutation_rate)) codeline->b = random_corenumber(max_number_size, 2);

    /* create instruction */
    codeline->in = OP(o, m, ma, mb);
}/*evolve_a_line()*/

/********************************************/

char* decode_line(const insn_t * codeline) {
    static char * line_ptr=0;
    static char line[256];
    int h;

    line_ptr = line;
    
    h = ((codeline->in >> opPOS) & opMASK) + TOK_OPCODE_START + 1;
    switch (h) {
    case TOK_DAT: strncpy(line_ptr, "dat", 3); break;
    case TOK_SPL: strncpy(line_ptr, "spl", 3); break;
    case TOK_MOV: strncpy(line_ptr, "mov", 3); break;
    case TOK_JMP: strncpy(line_ptr, "jmp", 3); break;
    case TOK_JMZ: strncpy(line_ptr, "jmz", 3); break;
    case TOK_JMN: strncpy(line_ptr, "jmn", 3); break;
    case TOK_ADD: strncpy(line_ptr, "add", 3); break;
    case TOK_SUB: strncpy(line_ptr, "sub", 3); break;
    case TOK_SEQ: strncpy(line_ptr, "seq", 3); break;
    case TOK_SNE: strncpy(line_ptr, "sne", 3); break;
    case TOK_MUL: strncpy(line_ptr, "mul", 3); break;
    case TOK_DIV: strncpy(line_ptr, "div", 3); break;
    case TOK_DJN: strncpy(line_ptr, "djn", 3); break;
    case TOK_SLT: strncpy(line_ptr, "slt", 3); break;
    case TOK_MOD: strncpy(line_ptr, "mod", 3); break;
    case TOK_NOP: strncpy(line_ptr, "nop", 3); break;
    default: printf("ARRG!! opcode [%d]\n", h); break;
    }

    line_ptr += 3;

    h = ((codeline->in >> moPOS) & moMASK) + TOK_MODIFIER_START + 1;
    switch(h) {
    case TOK_mF:  strncpy(line_ptr, ".f  ", 4); break;
    case TOK_mA:  strncpy(line_ptr, ".a  ", 4); break;
    case TOK_mB:  strncpy(line_ptr, ".b  ", 4); break;
    case TOK_mAB: strncpy(line_ptr, ".ab ", 4); break;
    case TOK_mBA: strncpy(line_ptr, ".ba ", 4); break;
    case TOK_mX:  strncpy(line_ptr, ".x  ", 4); break;
    case TOK_mI:  strncpy(line_ptr, ".i  ", 4); break;
    default: printf("ARRG!! modifier [%d]\n", h); break;
    }

    line_ptr += 4;

    h = ((codeline->in >> maPOS) & mMASK) + TOK_MODE_START + 1;
    switch(h) {
    case TOK_DIRECT:    strncpy(line_ptr, "$", 1); break;
    case TOK_IMMEDIATE: strncpy(line_ptr, "#", 1); break;
    case TOK_AINDIRECT: strncpy(line_ptr, "*", 1); break;
    case TOK_BINDIRECT: strncpy(line_ptr, "@", 1); break;
    case TOK_APREDEC:   strncpy(line_ptr, "{", 1); break;
    case TOK_BPREDEC:   strncpy(line_ptr, "<", 1); break;
    case TOK_APOSTINC:  strncpy(line_ptr, "}", 1); break;
    case TOK_BPOSTINC:  strncpy(line_ptr, ">", 1); break;
    default: printf("ARRG!! mode A [%d]\n", h); break;
    }

    line_ptr += 1;
    
    sprintf(line_ptr,"%7d, ", codeline->a);
    
    line_ptr += 9;
    
    h = ((codeline->in >> mbPOS) & mMASK) + TOK_MODE_START + 1;
    switch(h) {
    case TOK_DIRECT:    strncpy(line_ptr, "$", 1); break;
    case TOK_IMMEDIATE: strncpy(line_ptr, "#", 1); break;
    case TOK_AINDIRECT: strncpy(line_ptr, "*", 1); break;
    case TOK_BINDIRECT: strncpy(line_ptr, "@", 1); break;
    case TOK_APREDEC:   strncpy(line_ptr, "{", 1); break;
    case TOK_BPREDEC:   strncpy(line_ptr, "<", 1); break;
    case TOK_APOSTINC:  strncpy(line_ptr, "}", 1); break;
    case TOK_BPOSTINC:  strncpy(line_ptr, ">", 1); break;
    default: printf("ARRG!! mode B [%d]\n", h); break;
    }

    line_ptr += 1;

    sprintf(line_ptr,"%7d", codeline->b);
    
    line_ptr += 7;

    strcpy(line_ptr,"");

    return (line);
}
    
/********************************************/

void create_successor(const warrior_t * sw, warrior_t * tw) {
    int s_pos=0;
    int t_pos=0;
#ifdef DEBUG_YACE
    int i;
#endif

    if (sw == tw) {
	printf("Panic!! creat_successor called with Sourceaddress equal to Targetaddress!!\n");
	exit(1);
    }
    
    do {
	for (s_pos=0, t_pos=0; s_pos < sw->len; s_pos++, t_pos++) {
	    if (t_pos == max_instructions) {
		t_pos--;
		break;
	    }
	    if (propability(insertion_rate)) {		
		new_line_of_code(&tw->code[t_pos]);
		s_pos--;
		continue;
	    }
	    if (propability(removal_rate)) {
		t_pos--;
		continue;
	    }
	    
	    /* sw should be copied!! (does this work? Debug needed!) */
	    tw->code[t_pos] = sw->code[s_pos];
	    evolve_a_line(&tw->code[t_pos]);
	}
	tw->len = t_pos;
	tw->start = 0;
	tw->name = NULL;
    } while (tw->len > max_instructions || test_viable(tw)==0);

#ifdef DEBUG_YACE
    for (i=0; i < tw->len; i++) {
	printf("%s\n", decode_line(&tw->code[i]));
    }
    printf("\n");
#endif     
}/*create_successor()*/

/********************************************/

void fitness_standard(const int * warriors_selection, const warrior_t * warriors, warrior_score_t * warriors_score, int anz) {
    int i;
    int nr1=0;
    int nr2=0;

    /* clear scores */
    for (i=0; i<anz; i++) {
	warriors_score[i].score=0;
	warriors_score[i].fights=0;
    }
    
    for (i=0; i<anz; i+=2) {
	nr1 = warriors_selection[i];
	if (i+1 >= anz) nr2 = warriors_selection[0];
	else nr2 = warriors_selection[i+1];

	/* if popsize is uneven, last warrior also fights first */
	if (nr2 >= anz) nr2 = 0;
	
	mars(&warriors[nr1], &warriors[nr2], &warriors_score[nr1].score, &warriors_score[nr2].score, number_of_battles, OPT_cycle, OPT_minsep);
	warriors_score[nr1].fights += number_of_battles*2;
	warriors_score[nr2].fights += number_of_battles*2;
    }
}

/********************************************/

void fitness_testset(const int * warriors_selection, const warrior_t * warriors, warrior_score_t * warriors_score, int anz) {
    static warrior_t * testset = NULL;
    char file_name[256];
    int i, j, nr;
    int dummy;
    
    /* get warriors from directory 'hill' */
    if (testset == NULL) {
	testset = malloc(sizeof(warrior_t) * testset_size);
	for (i=0; i<testset_size; i++) {
	    sprintf(file_name,"..%stestset%s%d.red", path_symbol, path_symbol, i+1);
	    asm_fname(file_name, &testset[i], OPT_coresize);
	}
    }
    
    /* every selected warrior against testset */
    for (i=0; i<anz; i++) {
	nr = warriors_selection[i];
	warriors_score[nr].score=0;
	warriors_score[nr].fights=0;
	dummy=0;
	
	for (j=0; j<testset_size; j++) {
	    mars(&warriors[nr], &testset[j], &warriors_score[nr].score, &dummy, number_of_battles, OPT_cycle, OPT_minsep);
	    warriors_score[nr].fights += number_of_battles*2;
	}
#ifdef DEBUG_YACE
	printf("%5.1f, ", 100.0 * warriors_score[nr].score / warriors_score[nr].fights);
#endif
    }
}

/********************************************/

void reproduce_standard(const int * warriors_selection, warrior_t * warriors, warrior_score_t * warriors_score, int anz) {
    int i;
    int nr1;
    int nr2;
    double score1, score2;
    
    warrior_t * source;
    warrior_t * target;  

    for (i=0; i<anz; i+=2) {
	nr1 = warriors_selection[i];
	if (i+1 >= anz) nr2 = warriors_selection[0];
	else nr2 = warriors_selection[i+1];
	
	/* soure points to warrior with higher score */
	source = &warriors[nr1];
	target = &warriors[nr1];
	score1 = warriors_score[nr1].score / warriors_score[nr1].fights;
	score2 = warriors_score[nr2].score / warriors_score[nr2].fights;

	if (score1 < score2) {
	    source = &warriors[nr2];
	}
	else {
	    target = &warriors[nr2];
	}
	
	/* create successor */
	create_successor(source, target);
    }
}

/********************************************/

void straight_replace(const warrior_t * sw, warrior_t * tw) {
    *tw = *sw;
}

/********************************************/

void selection_random(int * warriors_selection, int anz) {
    int i;
    int nr;

    /* set numbers */
    for (i=0; i<anz; i++) warriors_selection[i] = i;

    /* create random numbers, every number only once */
    for (i=0; i<anz-1; i++) { /* anz-1: leave last number */
	if (warriors_selection[i] == i)	{
	    nr = random_number(anz-i-1)+i+1;
	    warriors_selection[i] = warriors_selection[nr];
	    warriors_selection[nr] = i;
	}
    }

    /*check last number */
    if (warriors_selection[anz-1] == anz-1) {
	nr = random_number(anz-1); /* replace with any number except last one (itself) */
	warriors_selection[anz-1] = warriors_selection[nr];
	warriors_selection[nr] = anz-1;
    }

#ifdef DEBUG_YACE
    printf("\nSelection:\n");
    for (i=0; i<anz; i++) printf("%2d ", warriors_selection[i]);
    printf("\n");    
#endif
}

/********************************************/

void mars(const warrior_t *war1, const warrior_t *war2, int *w1_score, int *w2_score, int rounds, unsigned long cycles, unsigned long distance) {
    int i=0, j=0, pos=0;
    int result=0;
    int start1=0;
    int start2=0;
        
    for (i=0; i<rounds; i++)
    {
        pos = distance + random_number(OPT_coresize - distance*2);
                
        for (j=0; j<2; j++)
        {
            /* coreclear */
            memset( core, 0, sizeof(insn_t) * OPT_coresize );

            start1=0;
            start2=0;
            if (j)  start1 = pos;
            else    start2 = pos;
                        
            memcpy(core + start1, war1->code, sizeof(insn_t) * war1->len);
            memcpy(core + start2, war2->code, sizeof(insn_t) * war2->len);
            result = sim( 2, war1->start + start1, war2->start + start2, cycles, NULL );

            /* one-on-one two warriors:
               0: warrior 1 won
               1: warrior 2 won
               2: tie
               -1: simulator panic attack -- something's gone wrong */
            switch(result)
            {
            case 0: *w1_score += 3; break;
            case 1: *w2_score += 3; break;
            case 2: *w1_score += 1; *w2_score += 1; break;
            case -1: printf("simulator panic attack!\n"); exit(1); break;
            }
        }/*for(0 --> 1)*/
    }/*for (0 --> rounds)*/
}/*mars()*/

/********************************************/

void status(const int * warriors_selection, const warrior_t * warriors, warrior_score_t * warriors_score, int anz, int current_cycle) {
    char file_name[256];
    FILE * warrior_file;
    double bestscore_adjusted=-1;
    double score_adjusted=0;
    int nr=-1;
    int i;

    /* calculate fitness values */
    printf("Status update, fighting against testset...\n");
    fitness_testset(warriors_selection, warriors, warriors_score, anz);
    printf(" Scores:\n");

    /* get nr of best warrior */
    for (i=0; i<anz; i++) {
	score_adjusted = 100.0 * warriors_score[i].score / warriors_score[i].fights;
	printf("%5.1f  ", score_adjusted);
	if (!((i+1) % 10)) printf("\n");
	if (score_adjusted > bestscore_adjusted) {
	    nr = i;
	    bestscore_adjusted = score_adjusted;
	}
    }
    bestscore_adjusted = 100.0 * warriors_score[nr].score / warriors_score[nr].fights;

    printf("\n");
    printf("Best one: nr. %d, score %5.1f\n", nr, bestscore_adjusted);
    
    /* print best warrior to file */
    sprintf(file_name, "..%sout%sev_%05.0f.red", path_symbol, path_symbol, bestscore_adjusted*100);
    if (0==(warrior_file=fopen(file_name,"w+"))) exit(0);	
    fprintf(warrior_file, ";redcode-94\n");
    fprintf(warrior_file, ";assert 1\n");
    fprintf(warrior_file, ";author Martin Ankerl <martinankerl@web.de>\n");
    fprintf(warrior_file, ";name Evolver %5d x %5ld\n", current_cycle, population_size);
    fprintf(warrior_file, ";strategy created with YACE 0.02\n");
    fprintf(warrior_file, ";strategy score against testset: %5.1f\n", bestscore_adjusted);

    for (i=0; i < warriors[nr].len; i++) {
	fprintf(warrior_file, "%s\n", decode_line(&warriors[nr].code[i]));
    }
    fprintf(warrior_file, "end 0\n");
    fclose(warrior_file);
}


void save_warriors(const warrior_t * warriors, int anz, int current_cycle) {
    FILE * warrior_file=0;
    char file_name[256];
    int i, j;

    printf("Saving population to disk\n");
    for (i=0; i<anz; i++) {
	/* print warriors to file */
	sprintf(file_name, "..%ssave%spop%05d.red", path_symbol, path_symbol, i);
	if (0==(warrior_file=fopen(file_name,"w+"))) {
	    printf("Can not write [%s] to disk!\n", file_name);
	}
	
	fprintf(warrior_file, ";redcode-94\n");
	fprintf(warrior_file, ";assert 1\n");
	fprintf(warrior_file, ";author Martin Ankerl <martinankerl@web.de>\n");
	fprintf(warrior_file, ";name Evolver %5d x %5ld\n", current_cycle, population_size);
	for (j=0; j < warriors[i].len; j++) {
	    fprintf(warrior_file, "%s\n", decode_line(&warriors[i].code[j]));
	}
	fprintf(warrior_file, "end 0\n");
	fclose(warrior_file);
    }
}



/********************************************/

int main() {
    warrior_t * warriors=0;
    warrior_score_t * warriors_score=0;
    int * warriors_selection=0;
    

    setup(&warriors, &warriors_score, &warriors_selection);

    /* if warriors were saved they are loaded otherwise randomly created */
    create_seed_population(warriors, population_size, startlength);
    
    for (current_cycle=0;; current_cycle++)
    {
	printf("cycle %5ld\n",current_cycle);

	/* create selection-list */
        selection_random(warriors_selection, population_size);

	/* run competion with selection list, create score-list */
	fitness_standard(warriors_selection, warriors, warriors_score, population_size);

	/* warriors_selection contains all warriors that should be killed and replaced */
	reproduce_standard(warriors_selection, warriors, warriors_score, population_size);

	/* some visualisation, print one warrior, ... */
	if (!(current_cycle % cycles_before_status) && current_cycle > 0) status(warriors_selection, warriors, warriors_score, population_size, current_cycle);

	/* Save population to disk */
	if (!(current_cycle % cycles_before_save) && current_cycle > 0) save_warriors(warriors, population_size, current_cycle);
    }
    return 0;
}
